/* $Id: f2ncalc.c,v 1.4 2007/02/20 14:48:06 dw Exp $ */
/* F_{2^n} interpreter */
/* Copyright (c) Damian Weber <dweber@htw-saarland.de> */
/* License: BSD-style
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "f2npoly.h"

#undef DEBUG

#define BUFFER_LENGTH 2048

/* read alpha, beta and operand from input */
/* return 1 = okay                         */
/* return 0 = any error                    */

int read_operands(char *buffer, char *op, f2poly_t  *alphap, f2poly_t *betap, ulong *ep)
{
	char *bp,*pp;
	char *ps1,*pe1,*ps2,*pe2;
	unsigned int i;
	char functions[5][10]=
	{
		"unknown",
		"gencheck",
		"findgen",
		"sqrt",
		"help"
	};

	bp=buffer;

	if (buffer[0]=='(') /* expression is (f(X))op(g(X)) */
	{
		ps1=buffer+1;
		pe1=strchr(ps1,')');
		if (!pe1)
		{
			puts("')' of 1st operand expected");
			return 0;
		}
		*op=*(pe1+1);
		/*printf("operand character %c\n",*op); */
		if (!strchr("+*/^",*op))
		{
			puts("no operand found");
			return 0;
		}
		*pe1=0; /* end of string 1st operand */
		to_f2poly(alphap,ps1);
			
		if (*op=='^')
			sscanf(pe1+2,"%lu",ep);
		else
		{
			ps2=pe1+2;
			if (*ps2!='(')
			{
				puts("'(' of 2nd operand expected");
				return 0;
			}
			pe2=strchr(ps2,')');
			if (!pe1)
			{
				puts("')' of 2nd operand expected");
				return 0;
			}
			*pe2=0; /* end of string 2nd operand */
			to_f2poly(betap,ps2+1);
		}
	}
	else
	{
		/* functions */
		for (i=1;i<sizeof(functions)/sizeof(functions[0]);i++)
		{
			if (!strncmp(buffer,functions[i],strlen(functions[i])))
			{
				bp+=strlen(functions[i]);
				if (*bp != '(') 
				{
					printf("not a bracket following %s\n",functions[i] );
					return 0;
				}
				bp++;
				pp=bp;
				while (*pp && (*pp!=')')) pp++;
				*pp=0;

				to_f2poly(alphap,bp);

				*pp=')';
				*op=i;
				return 1;
			}
		}

		if (buffer[0]!='\n') /* on empty input line, do nothing*/
		{
			puts("syntax error, try \"help()\"");
			return 0; 
		}
		else
		{
			*op='#';
			return 1;
		}
	}
	return 1;
}

void f2n_interpreter(const f2poly_t *fp)
{
	const char *prompt = "f2ncalc> ";
	char buffer[BUFFER_LENGTH], op;
	f2poly_t alpha, beta, gamma,dummy;

	ulong e;
	ulong pn;
	ulong primes[8*sizeof(ulong)];
	int expo[8*sizeof(ulong)];
	int line;
	int okay;
	int k;
	int nf;
	int suppress_output;
	int rc_gen;

	init_poly(&alpha);
	init_poly(&beta);
	init_poly(&gamma);
	init_poly(&dummy);

	pn=(ulong)1<<fp->n;
	ul_factor(pn-1,primes,expo,&nf);

	printf("2^n-1=%lu",pn-1);
	if (!nf)
		puts(" (prime)");
	else
	{
		printf("=%lu",primes[0]);
		if (expo[0]>1)
			printf("^%d",expo[0]);
		for (k=1;k<=nf;k++)
		{
			if (expo[k]>1)
				printf("*%lu^%d",primes[k],expo[k]);
			else
				printf("*%lu",primes[k]);
		}
	}
	puts("");
	line=0;

	while (1)
	{
		printf("%s",prompt);
		fflush(stdout);
		if (!fgets(buffer, BUFFER_LENGTH,stdin))
		{
			suppress_output=1;
			break;
		}
		if (buffer[0] == 'q')
		{
			suppress_output=1;
			break;
		}

		suppress_output=0;

		okay = read_operands(buffer, &op, &alpha, &beta, &e);

		if (okay)
		{
#ifdef DEBUG
			to_string(buffer,&alpha);
			printf("alpha=%s\n",buffer);
			if (op!='^')
			{
				to_string(buffer,&beta);
				printf("beta =%s\n",buffer);
			}
			else
				printf("e=%lu\n",e);
			
			printf("op=%c\n",op);
#endif
			switch (op)
			{
			    case '^':
				fast_power_mod(&gamma,&alpha, e, fp);
				break;
			    case '+':
				add(&gamma,&alpha,&beta);
				div_rem(&dummy,&gamma,&gamma,fp);
				break;
			    case '*':
				mul(&gamma,&alpha,&beta);
				div_rem(&dummy,&gamma,&gamma,fp);
				break;
			    case '/':
				fast_power_mod(&gamma,&beta, pn-2, fp);
				mul(&gamma,&alpha,&gamma);
				div_rem(&dummy,&gamma,&gamma,fp);
				break;
			    case 1: /* gencheck()*/
				rc_gen= is_generator(fp, &alpha, pn-1, primes, nf);
				to_string(buffer,&alpha);
				printf("%s is %sa generator\n",buffer,(rc_gen)? "":"not ");
				suppress_output=1;
				break;
			    case 2:
				generator_search(fp, &gamma);
				to_string(buffer,&gamma);
				printf("%s\n",buffer);
				suppress_output=1;
				
				break;
			    case 3:
				fast_power_mod(&gamma,&alpha, (pn+1)/2, fp);
				div_rem(&dummy,&gamma,&gamma,fp);
				break;
			    case 4:
				puts("HELP: " );
				puts("arithmetic operations +,*,/,^" );
				puts("  try \"(X+1)+(X^2+1)\" " );
				puts("functions:            " );
				puts("   - gencheck" );
				puts("   - findgen  " );
				puts("   - sqrt     " );
				suppress_output=1;
				break;
			    case '#': /* empty input */
				suppress_output=1;
				break;
			    default:
				puts("error: wrong operand '" );
				okay = 0;
			}
		}

		if (okay && !suppress_output)
		{
			to_string(buffer,&gamma);
			printf("out(%d)=%s\n",++line,buffer);
		}
	}
	puts("");
	puts("Thanks for using f2ncalc. Good bye.");
}



int main(int argc, char **argv)
{
	char buffer[1000];
	f2poly_t f,g;
	long n;
	int input_ok;

	char *welcome="f2ncalc - A simple interpreter for doing calculations in finite fields F_{2^n}.\n";

	if (argc <= 1)
	{
		printf("%s:\n usage %s n\n",argv[0],argv[0]);
		printf(" or    %s poly\n",argv[0]);
		return 1;
	}

	init_poly(&f);
	init_poly(&g);
	
	puts(welcome);

	input_ok=0;
	if (isdigit(argv[1][0]))
	{
		n = atol(argv[1]);
		if (n<=1)
		{
			printf("%s: use n>1, please",argv[0]);
			return 1;
		}
		irred_search(&f,n,0);
		to_string(buffer,&f);
		printf("field polynomial %s\n",buffer);
		input_ok=1;
	}

	if (argv[1][0]=='X')
	{
		to_f2poly(&f,argv[1]);
		if (!irred_test(&f,&g))
		{
			printf("the given poly %s is not irreducible\n",argv[1]);
			to_string(buffer,&g);
			printf("a factor is %s, aborting\n",buffer);
			return 1;
		}
	}

	f2n_interpreter(&f);

	return 0;
}
